==> irc-transport-0.1b-1.40.linux.tar.gz: irc-transport-0.1b-1.40/src/irc_in.c <==
/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>
#include <sys/utsname.h>


irc_session irc_new_session(irctrans_data ircdata, jpacket jp)
{
    irc_session new;
    pool        p;

    log_debug(ZONE, "EXTRA_DEBUG: creating new session for jpacket %s", xmlnode2str(jp->x));
    p             = pool_new();
    new           = pmalloco(p, sizeof(_irc_session));
    new->j        = jid_new(p, jid_full(jp->from));
    new->id       = spools(p, jid_full(jp->from), "/", irc_get_server(p, jid_full(jp->to)), p);
    new->p        = p;
    new->server   = irc_get_server(p, jid_full(jp->to));
    new->channels = xhash_new(101);
    new->data     = (void*)ircdata;
    new->nick     = pstrdup(p, jp->to->user);


    return new;
}

void irc_parse_server(irctrans_data ircdata, irc_session s, jpacket jp)
{

    log_debug(ZONE, "EXTRA_DEBUG: server parse of jpacket");
    if(jp->type != JPACKET_IQ)
    {
        xmlnode_free(jp->x);
        return;
    }

    if(jpacket_subtype(jp) != JPACKET__GET)
    {
        jutil_error(jp->x, TERROR_NOTALLOWED);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    if(NSCHECK(jp->iq,NS_TIME))
    {
        time_t t;
        char   *str;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_TIME);
        jpacket_reset(jp);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"utc"),jutil_timestamp(),-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"tz"),tzname[0],-1);
        
        /* create nice display time */
        t = time(NULL);
        str = ctime(&t);
        str[strlen(str) - 1] = '\0'; /* cut off newline */
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"display"),str,-1);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
   
    if(NSCHECK(jp->iq,NS_VERSION))
    {
        struct utsname un;
        xmlnode x;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_VERSION);
        jpacket_reset(jp);
 
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"name"),"conference",-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"version"),VERSION,-1);
 
        uname(&un);
        x = xmlnode_insert_tag(jp->iq,"os");
        xmlnode_insert_cdata(x,un.sysname,-1);
        xmlnode_insert_cdata(x," ",1);
        xmlnode_insert_cdata(x,un.release,-1);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    if(NSCHECK(jp->iq,NS_BROWSE))
    {
        jutil_iqresult(jp->x);
        xmlnode_insert_node(jp->x, ircdata->browse);
        jpacket_reset(jp);
        xmlnode_put_attrib(jp->iq,"type","irc");
        xmlnode_put_attrib(jp->iq,"name",xmlnode_get_tag_data(ircdata->cfg,"vCard/FN")); /* pull name from the server vCard */
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    if(NSCHECK(jp->iq,NS_LAST))
    {
        time_t start;
        char nstr[20];
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_LAST);
        jpacket_reset(jp);
       
        start = time(NULL) - ircdata->start;
        sprintf(nstr,"%d",start);
        xmlnode_put_attrib(jp->iq,"seconds",nstr);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }

    if(NSCHECK(jp->iq,NS_VCARD))
    {
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"vCard"),"xmlns",NS_VCARD);
        jpacket_reset(jp);

        xmlnode_insert_node(jp->iq,xmlnode_get_firstchild(xmlnode_get_tag(ircdata->cfg,"vCard")));
        deliver(dpacket_new(jp->x),NULL);
        return;
    }


    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x),NULL);
}

void irc_parser(void *arg)
{
    jpacket jp = (jpacket)arg;
    irctrans_data ircdata = jp->aux1;
    irc_session s;
    
    log_debug(ZONE, "EXTRA_DEBUG: irc_parser handling incoming packet from MTQ"); 
    if(jpacket_subtype(jp) == JPACKET__ERROR)
    {
        xmlnode_free(jp->x);
        return;
    }

    s = (irc_session)xhash_get(ircdata->sessions, spools(jp->p, jid_full(jp->from), "/", irc_get_server(jp->p, jid_full(jp->to)), jp->p));

    log_debug(ZONE, "EXTRA_DEBUG: found session %X", s);
    /* branch off server packets */
    if(jp->to->user == NULL)
    {
        irc_parse_server(ircdata, s, jp);
        return;
    }
    
    if(jp->type == JPACKET_IQ && NSCHECK(jp->iq, NS_CONFERENCE) && jpacket_subtype(jp) == JPACKET__GET)
    {
        irc_session s = irc_new_session(ircdata, jp);
        log_debug(ZONE, "EXTRA_DEBUG: IQ GET starting a get thread");
        s->data = (void*)ircdata;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"), "xmlns", NS_CONFERENCE);
        jpacket_reset(jp);
        xmlnode_insert_tag(jp->iq, "nick");
        jp->aux1 = (void*)s;
        log_debug(ZONE, "calling mio_connect(%s, %d)", s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667));
        mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667), irc_check_iq_get, (void*)jp, 15, MIO_CONNECT_RAW);
        return;
    }

    if(s == NULL)
    {
        if(*jp->x->name != 'p' || jutil_priority(jp->x) < 0)
        {
            if(jpacket_subtype(jp) == JPACKET__ERROR)
            {
    log_debug(ZONE, "EXTRA_DEBUG: freeing error packet");
                xmlnode_free(jp->x);
                return;
            }

            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }

        /* start a new irc session */
        s = irc_new_session(ircdata, jp);

        if(s == NULL)
        {
            jutil_error(jp->x, TERROR_NOTALLOWED);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        xhash_put(ircdata->sessions, s->id, (void*)s);
        xmlnode_free(jp->x);
        return;
    }

    if(jp->type == JPACKET_IQ)
    {
        if(NSCHECK(jp->iq, NS_CONFERENCE) && jpacket_subtype(jp) == JPACKET__SET)
        {
            jp->aux1 = (void*)s;
            s->data = (void*)ircdata;
    log_debug(ZONE, "EXTRA_DEBUG: IQ SET starting a regular session thread");
            mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->to)), 6667), irc_parse_phase_zero, (void*)jp, 10, MIO_CONNECT_RAW);
            return;
        }

        /* we don't accept this */
        jutil_error(jp->x, TERROR_NOTIMPL);
        deliver(dpacket_new(jp->x), NULL);
        return;
    }

    if(jp->type == JPACKET_PRESENCE && jpacket_subtype(jp) == JPACKET__UNAVAILABLE)
    {
        /* end session */
        xhash_zap(ircdata->sessions, s->id);
        mio_close(s->m);
        return;
    }
    else if(jp->type == JPACKET_PRESENCE)
    {
        xmlnode_free(jp->x);
        return;
    }

    log_debug(ZONE, "EXTRA_DEBUG: Messaging is not finished");
    /* all that's left are messages to the group */

    if(jp->type == JPACKET_MESSAGE)
    {
        user u;
        char *msg;
        chans channel = xhash_get(s->channels, spools(jp->p, jp->to->user, "@", jp->to->server, jp->p));
        if(channel == NULL)
        {
            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        if(jp->to->resource != NULL) /* private message */
        {
            jutil_error(jp->x, TERROR_NOTIMPL);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        msg = UTF8_to_str(jp->p, xmlnode_get_data(xmlnode_get_tag(jp->x, "body")));

        if(strlen(msg) > 450)
            msg[450] = '\0';

        
        /*
        * this if block handles the packet.. 
        * each if section is responsible for freeing
        * the packet in it's own way
        */
        if(strncasecmp(msg, "/me ", 4) == 0)
        {
            /* this should be an action */
            jutil_tofrom(jp->x);
            msg += 4;
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            jpacket_reset(jp);

            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :\001ACTION ", msg, "\001\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/say ", 5) == 0)
        {
            jutil_tofrom(jp->x);
            msg += 5;
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), msg, -1);

            jpacket_reset(jp);
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/raw ", 5) == 0)
        {
            msg += 5;
            mio_write(s->m, NULL, spools(jp->p, msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg,"/whowas ",7)==0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOWAS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/whois ", 7) == 0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOIS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/msg ", 5) == 0 || strncasecmp(msg, "/query ", 7) == 0)
        {
            char *nick, *sendmsg;
            if(strncasecmp(msg, "/msg ", 5) == 0)
                msg += 5;
            else
                msg += 7;

            if(msg != NULL && msg[0] != '\0')
            {
                nick = msg;
                msg = strchr(msg,' ');
                if(msg != NULL)
                {
                    msg[0] = '\0';
                    msg++;
                    /* strip the @ and + from the user */
                    if(nick[0] == '@' || nick[0] == '+') nick++;
                    mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", nick, " :", msg, "\n", jp->p), -1);
                }
            }
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), ">>> PRIVMSG SENT", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/mode ", 6) == 0)
        {
            msg += 6;
            mio_write(s->m, NULL, spools(jp->p, "MODE ", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(msg[0] == '/')
        {
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), "/ COMMAND NOT SUPPORTED", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else /* standard message */
        {
            jutil_tofrom(jp->x);
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            jpacket_reset(jp);
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        return;
    }

    /* dunno what this is */
    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x), NULL);
}

==> irc-transport-0.1.1-1.4.linux.tar.gz: irc-transport-0.1.1-1.40/src/irc_in.c <==
/*
IRC-Transport main.c: Main program loop
Copyright (C) 1999-2000 Keith Minkler <tsbandit@jabber.org>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

You can reach The author at tsbandit@jabber.org (email or jid)
*/

#include <irctrans.h>
#include <sys/utsname.h>


irc_session irc_new_session(irctrans_data ircdata, jpacket jp)
{
    irc_session new;
    pool        p;

    p             = pool_new();
    new           = pmalloco(p, sizeof(_irc_session));
    new->j        = jid_new(p, jid_full(jp->from));
    jp->from->next = (jid)ircdata;
    new->server   = irc_get_server(p, jp->from, jid_full(jp->to));
    log_debug(ZONE, "found server name: %s", new->server);
    new->id       = spools(p, jid_full(jp->from), "/", new->server, p);
    new->p        = p;
    new->channels = xhash_new(101);
    new->data     = (void*)ircdata;
    new->nick     = pstrdup(p, jp->to->user);


    return new;
}

void irc_parse_server(irctrans_data ircdata, irc_session s, jpacket jp)
{

    if(jp->type != JPACKET_IQ)
    {
        xmlnode_free(jp->x);
        return;
    }
    else if(jpacket_subtype(jp) != JPACKET__GET && !NSCHECK(jp->iq, NS_REGISTER))
    {
        jutil_error(jp->x, TERROR_NOTALLOWED);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_TIME))
    {
        time_t t;
        char   *str;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_TIME);
        jpacket_reset(jp);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"utc"),jutil_timestamp(),-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"tz"),tzname[0],-1);
        
        /* create nice display time */
        t = time(NULL);
        str = ctime(&t);
        str[strlen(str) - 1] = '\0'; /* cut off newline */
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"display"),str,-1);
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_VERSION))
    {
        struct utsname un;
        xmlnode x;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_VERSION);
        jpacket_reset(jp);
 
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"name"),"conference",-1);
        xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq,"version"),VERSION,-1);
 
        uname(&un);
        x = xmlnode_insert_tag(jp->iq,"os");
        xmlnode_insert_cdata(x,un.sysname,-1);
        xmlnode_insert_cdata(x," ",1);
        xmlnode_insert_cdata(x,un.release,-1);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_BROWSE))
    {
        irc_session s;

        if(jp->to->resource == NULL)
        {
            s = irc_new_session(ircdata, jp);
        
            jutil_iqresult(jp->x);

            if(jp->to->user == NULL) /* browse request to a server */
                xmlnode_insert_node(jp->x, ircdata->browse);

            jpacket_reset(jp);
    
            s->data = (void*)ircdata;
            jp->aux1 = (void*)s;
            log_debug(ZONE, "calling mio_connect(%s, %d)", s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667));
            mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667), irc_check_browse_get, (void*)jp, 15, MIO_CONNECT_RAW);
            return;
        }
    }
    else if(NSCHECK(jp->iq,NS_LAST))
    {
        time_t start;
        char nstr[20];
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"),"xmlns",NS_LAST);
        jpacket_reset(jp);
       
        start = time(NULL) - ircdata->start;
        sprintf(nstr,"%d",start);
        xmlnode_put_attrib(jp->iq,"seconds",nstr);
   
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_VCARD))
    {
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"vCard"),"xmlns",NS_VCARD);
        jpacket_reset(jp);

        xmlnode_insert_node(jp->iq,xmlnode_get_firstchild(xmlnode_get_tag(ircdata->cfg,"vCard")));
        deliver(dpacket_new(jp->x),NULL);
        return;
    }
    else if(NSCHECK(jp->iq,NS_REGISTER))
    {
        if(jpacket_subtype(jp) == JPACKET__GET)
        {
            jid temp = jid_new(jp->p, jid_full(jp->from));
            jid j;
            xmlnode x;
            char *servername;

            jid_set(temp, NULL, JID_RESOURCE);
            j = jid_new(jp->p, spools(jp->p, shahash(jid_full(temp)), "@", ircdata->i->id, jp->p));
            x = xdb_get(ircdata->xc, j, "irc:default:server");
            x = xmlnode_get_firstchild(x);
            servername = xmlnode_get_data(x) ? xmlnode_get_data(x) : xmlnode_get_tag_data(ircdata->cfg, "default/server");

            jutil_iqresult(jp->x);
            xmlnode_put_attrib(xmlnode_insert_tag(jp->x, "query"), "xmlns", NS_REGISTER);
            jpacket_reset(jp);

            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "instructions"), "Enter the name of the default server you wish to use when you log into a group on IRC Transport.  You may log into another server at any time using the format [#|.]Channel[\%server_name] as the name of the group you are joining", -1);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "misc"), servername, -1);
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->iq, "key"), jutil_regkey(NULL, jid_full(jp->to)), -1);
            if(x != NULL) xmlnode_insert_tag(jp->iq, "registered");
            deliver(dpacket_new(jp->x), NULL);
            xmlnode_free(x);
            return;
        }
        else if(jpacket_subtype(jp) == JPACKET__SET)
        {
            xmlnode x = xmlnode_new_tag("server");
            jid temp = jid_new(jp->p, jid_full(jp->from));
            jid j;
            char *servername;

            jid_set(temp, NULL, JID_RESOURCE);
            j = jid_new(jp->p, spools(jp->p, shahash(jid_full(temp)), "@", ircdata->i->id, jp->p));

            /* verify data */
            if(xmlnode_get_tag(jp->iq, "remove") != NULL)
            {
                xdb_act(ircdata->xc, j, "irc:default:server","insert", "server", x);
                xmlnode_free(x);
                jutil_iqresult(jp->x);
                deliver(dpacket_new(jp->x), NULL);
                return;
            }

            servername = xmlnode_get_tag_data(jp->iq, "misc") ? xmlnode_get_tag_data(jp->iq, "misc") : xmlnode_get_tag_data(ircdata->cfg, "default/server");
            xmlnode_insert_cdata(x, servername, -1);
            xdb_act(ircdata->xc, j, "irc:default:server", "insert", "server", x);
            servername = x ? xmlnode_get_data(x) : xmlnode_get_tag_data(ircdata->cfg, "default/server");

            xmlnode_free(x);
            jutil_iqresult(jp->x);
            deliver(dpacket_new(jp->x), NULL);
            return;

        }
    }


    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x),NULL);
}

void irc_parser(void *arg)
{
    jpacket jp = (jpacket)arg;
    irctrans_data ircdata = jp->aux1;
    irc_session s;
    
    if(jpacket_subtype(jp) == JPACKET__ERROR)
    {
        xmlnode_free(jp->x);
        return;
    }

    jp->from->next = (jid)ircdata;
    s = (irc_session)xhash_get(ircdata->sessions, spools(jp->p, jid_full(jp->from), "/", irc_get_server(jp->p, jp->from, jid_full(jp->to)), jp->p));

    /* branch off server packets */
    if(jp->to->user == NULL || NSCHECK(jp->iq, NS_BROWSE))
    {
        irc_parse_server(ircdata, s, jp);
        return;
    }
    
    if(jp->type == JPACKET_IQ && NSCHECK(jp->iq, NS_CONFERENCE) && jpacket_subtype(jp) == JPACKET__GET)
    {
        irc_session s = irc_new_session(ircdata, jp);
        s->data = (void*)ircdata;
        jutil_iqresult(jp->x);
        xmlnode_put_attrib(xmlnode_insert_tag(jp->x,"query"), "xmlns", NS_CONFERENCE);
        jpacket_reset(jp);
        xmlnode_insert_tag(jp->iq, "nick");
        jp->aux1 = (void*)s;
        log_debug(ZONE, "calling mio_connect(%s, %d)", s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667));
        mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->from)), 6667), irc_check_iq_get, (void*)jp, 15, MIO_CONNECT_RAW);
        return;
    }

    if(s == NULL)
    {
        if(*jp->x->name != 'p' || jutil_priority(jp->x) < 0)
        {
            if(jpacket_subtype(jp) == JPACKET__ERROR)
            {
                xmlnode_free(jp->x);
                return;
            }

            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x),NULL);
            return;
        }

        /* start a new irc session */
        s = irc_new_session(ircdata, jp);
        log_notice(jp->to->server, "Initial Session Started for user %s for channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);

        if(s == NULL)
        {
            jutil_error(jp->x, TERROR_NOTALLOWED);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        xhash_put(ircdata->sessions, s->id, (void*)s);
        xmlnode_free(jp->x);
        return;
    }

    if(jp->type == JPACKET_IQ)
    {
        if(NSCHECK(jp->iq, NS_CONFERENCE) && jpacket_subtype(jp) == JPACKET__SET)
        {
            char *chan = make_lowercase(irc_get_channel(jp->p, jid_full(jp->to)));
            chans channel = xhash_get(s->channels, chan);

            if(s->m == NULL)
            { /* we don't have a connection to the server yet */
                jp->aux1 = (void*)s;
                s->data = (void*)ircdata;
                log_notice(jp->to->server, "Bootstrapping Connection for user %s: channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);
                mio_connect(s->server, j_atoi(irc_get_port(s->p, jid_full(jp->to)), 6667), irc_parse_phase_zero, (void*)jp, 10, MIO_CONNECT_RAW);
                return;
            }

            if(channel == NULL)
            { /* startup a new channel for this session */

                mio_reset(s->m, irc_parse_phase_one, jp);
                mio_write(s->m, NULL, spools(jp->p, "JOIN ", irc_get_channel(jp->p, jid_full(jp->to)), " ", xmlnode_get_tag_data(jp->x, "query/password"), "\n", jp->p), -1);
                return;
            }

            /* XXX channel is not null.. this must be a nick change */
            if(xmlnode_get_tag_data(jp->iq, "nick") == NULL)
            {
                jutil_error(jp->x, TERROR_NOTALLOWED);
                deliver(dpacket_new(jp->x), NULL);
                return;
            }
            s->aux1 = (void*)jp;
            mio_write(s->m, NULL, spools(jp->p, "NICK :", xmlnode_get_tag_data(jp->iq, "nick"), "\n", jp->p), -1);
            xmlnode_hide(xmlnode_get_tag(jp->iq, "nick"));
            return;
        }

        /* we don't accept this */
        jutil_error(jp->x, TERROR_NOTIMPL);
        deliver(dpacket_new(jp->x), NULL);
        return;
    }

    if(jp->type == JPACKET_PRESENCE && jpacket_subtype(jp) == JPACKET__UNAVAILABLE)
    {
        /* end session */
        log_notice(jp->to->server, "Initial Session Shut Down for user %s for channel %s on server %s", jid_full(jp->from), irc_get_channel(jp->p, jid_full(jp->to)), s->server);
        xhash_zap(ircdata->sessions, s->id);
        mio_close(s->m);
        return;
    }
    else if(jp->type == JPACKET_PRESENCE)
    {
        xmlnode_free(jp->x);
        return;
    }

    /* all that's left are messages to the group */

    if(jp->type == JPACKET_MESSAGE)
    {
        user u;
        char *msg;
        chans channel = xhash_get(s->channels, spools(jp->p, jp->to->user, "@", jp->to->server, jp->p));
        if(channel == NULL)
        {
            jutil_error(jp->x, TERROR_NOTFOUND);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        if(jp->to->resource != NULL) /* private message */
        {
            jutil_error(jp->x, TERROR_NOTIMPL);
            deliver(dpacket_new(jp->x), NULL);
            return;
        }

        msg = UTF8_to_str(jp->p, xmlnode_get_data(xmlnode_get_tag(jp->x, "body")));

        if(strlen(msg) > 450)
            msg[450] = '\0';

        
        /*
        * this if block handles the packet.. 
        * each if section is responsible for freeing
        * the packet in it's own way
        */
        if(strncasecmp(msg, "/me ", 4) == 0)
        {
            /* this should be an action */
            jutil_tofrom(jp->x);
            msg += 4;
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            jpacket_reset(jp);

            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :\001ACTION ", msg, "\001\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/say ", 5) == 0)
        {
            jutil_tofrom(jp->x);
            msg += 5;
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), msg, -1);

            jpacket_reset(jp);
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/raw ", 5) == 0)
        {
            msg += 5;
            mio_write(s->m, NULL, spools(jp->p, msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg,"/whowas ",7)==0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOWAS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/whois ", 7) == 0)
        {
            msg += 7;
            mio_write(s->m, NULL, spools(jp->p, "WHOIS :", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(strncasecmp(msg, "/msg ", 5) == 0 || strncasecmp(msg, "/query ", 7) == 0)
        {
            char *nick, *sendmsg;
            if(strncasecmp(msg, "/msg ", 5) == 0)
                msg += 5;
            else
                msg += 7;

            if(msg != NULL && msg[0] != '\0')
            {
                nick = msg;
                msg = strchr(msg,' ');
                if(msg != NULL)
                {
                    msg[0] = '\0';
                    msg++;
                    /* strip the @ and + from the user */
                    if(nick[0] == '@' || nick[0] == '+') nick++;
                    mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", nick, " :", msg, "\n", jp->p), -1);
                }
            }
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), ">>> PRIVMSG SENT", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else if(strncasecmp(msg, "/mode ", 6) == 0)
        {
            msg += 6;
            mio_write(s->m, NULL, spools(jp->p, "MODE ", msg, "\n", jp->p), -1);
            xmlnode_free(jp->x);
        }
        else if(msg[0] == '/')
        {
            jutil_tofrom(jp->x);
            xmlnode_hide(xmlnode_get_tag(jp->x, "body"));
            xmlnode_insert_cdata(xmlnode_insert_tag(jp->x, "body"), "/ COMMAND NOT SUPPORTED", -1);
            xmlnode_put_attrib(jp->x, "from", spools(jp->p, jp->from->user, "@", jp->from->server, jp->p));
            deliver(dpacket_new(jp->x), NULL);
        }
        else /* standard message */
        {
            jutil_tofrom(jp->x);
            xmlnode_put_attrib(jp->x, "from", jid_full(s->nick_id));
            jpacket_reset(jp);
            mio_write(s->m, NULL, spools(jp->p, "PRIVMSG ", irc_get_channel(jp->p, jid_full(jp->from)), " :", msg, "\n", jp->p), -1);
            deliver(dpacket_new(jp->x), NULL);
        }
        return;
    }

    /* dunno what this is */
    jutil_error(jp->x, TERROR_NOTIMPL);
    deliver(dpacket_new(jp->x), NULL);
}
